# Alea version 1.0.0
# DiscreteEventSimulator.py
# by Antoine Rauzy
# Copyrights (c) 2020 AltaRica Association
#
# This file is part of Alea, a pedagogical implementation of stochastic discrete event systems.
# Alea is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Alea is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Alea.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Simulators
#   2.1. System
#   2.2. Executions
#   2.3. Indicators
# 3. Schedules
# 4. Events
# 5. Indicators
#   5.1. Base Indicators
#   5.2. Firing Counts
#   5.3. Transition Firing Counts
#   5.4. Transition First Firing Dates
#   5.5. Predicate Satisfaction Counts
#   5.6. Predicate First Satisfaction Dates
#   5.7. Predicate Satisfaction Time
#   5.8. Reward Minimum Values
#   5.9. Reward Maximum Values
#   5.10. Reward Mean Values

# 1. Required modules
# -------------------

import sys
import random

from Alea.DiscreteEventSystems import *

# 2. Simulators
# -------------

class DiscreteEventSimulator:
  def __init__(self):
    self.system = None
    self.currentState = None
    self.schedule = Schedule()
    self.missionTime = 0
    self.currentDate = 0
    self.indicators = []

# 2.1. System
# -----------

  def GetSystem(self):
    return self.system

  def SetSystem(self, system):
    self.system = system
    self.currentState = self.system.NewState()
    self.schedule.SetSystem(system)
    self.indicators.clear()

# 2.2. Executions
# ---------------

  def MonteCarloSimulation(self, missionTime, numberOfTrials):
    if self.system==None:
      return
    self.ResetIndicators()
    for trial in range(0, numberOfTrials):
      self.SimulateExecution(missionTime)

  def SimulateExecution(self, missionTime):
    if self.system==None:
      return
    self.missionTime = missionTime
    self.currentDate = 0
    self.currentState.Reset()
    self.schedule.Restart()
    self.UpdateSchedule()
    self.ResetIndicatorsAtBeginningOfExecution()
    while self.currentDate<=self.missionTime and self.schedule.GetNumberOfScheduledEvents()!=0:
      event = self.SelectCandidateEvent()
      if event==None:
        break
      transition = event.GetTransition()
      firingDate = event.GetFiringDate()
      if firingDate>self.missionTime:
        break
      self.schedule.UnscheduleEvent(event)
      if transition.GetPolicy()==Transition.MEMORY:
        event.SetRemainingTime(0)
      self.currentDate = firingDate
      self.system.FireTransition(transition, self.currentState)
      # print("fire transition {0:s} at {1:f}".format(transition.GetName(), self.currentDate))
      self.UpdateSchedule()
      self.UpdateIndicatorsAfterTransitionFiring(transition, firingDate)
    self.UpdateIndicatorsAtEndOfExecution(self.missionTime)

  def SelectCandidateEvent(self):
    candidateEvents = self.schedule.GetCandidateEvents()
    if len(candidateEvents)==0:
      return None
    if len(candidateEvents)==1:
      return candidateEvents[0]
    sumOfWeights = 0
    for candidateEvent in candidateEvents:
      candidateTransition = candidateEvent.GetTransition()
      weight = candidateTransition.CalculateWeight(self.currentState)
      sumOfWeights += weight
    if sumOfWeights<=0:
      return None
    z = random.random()
    sumOfWeights2 = 0
    for candidateEvent in candidateEvents:
      candidateTransition = candidateEvent.GetTransition()
      weight = candidateTransition.CalculateWeight(self.currentState)
      sumOfWeights2 += weight
      if z<=sumOfWeights2/sumOfWeights:
        return candidateEvent
    return None

  def UpdateSchedule(self):
    # print("++UpdateSchedule at {0:f}".format(self.currentDate))
    for event in self.schedule.GetEvents():
      transition = event.GetTransition()
      if event.GetFiringDate()==Delay.INFINITY:
        if self.system.IsTransitionEnabled(transition, self.currentState):
          schedulingDate = self.currentDate
          if transition.GetPolicy()==Transition.MEMORY and event.GetRemainingTime()>0:
            delay = event.GetRemainingTime()
            # print("reschedule transition {0:s} at {1:f} for {2:f}".format(transition.GetName(), schedulingDate, delay))
          else:
            delay = self.system.CalculateDelayOfTransition(transition, self.currentState)
          firingDate = schedulingDate + delay
          self.schedule.ScheduleEvent(event, schedulingDate, firingDate)
          # print("schedule transition {0:s} at {1:f} for {2:f}".format(transition.GetName(), schedulingDate, firingDate))
      else:
        if not self.system.IsTransitionEnabled(transition, self.currentState):
          if transition.GetPolicy()==Transition.MEMORY:
            event.SetRemainingTime(event.GetFiringDate() - self.currentDate)
          self.schedule.UnscheduleEvent(event)
          # print("unschedule transition {0:s} at {1:f}".format(transition.GetName(), self.currentDate))
    # print("--UpdateSchedule")

# 2.3. Indicators
# ---------------

  def GetIndicators(self):
    return self.indicators

  def NewFiringCount(self, name, recordValues):
    indicator = FiringCount(name, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewTransitionFiringCount(self, name, transition, recordValues):
    indicator = TransitionFiringCount(name, transition, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewTransitionFirstFiringDate(self, name, transition, recordValues):
    indicator = TransitionFirstFiringDate(name, transition, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewPredicateSatisfactionCount(self, name, predicate, recordValues):
    indicator = PredicateSatisfactionCount(name, predicate, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewPredicateFirstSatisfactionDate(self, name, predicate, recordValues):
    indicator = PredicateFirstSatisfactionDate(name, predicate, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewPredicateSatisfactionTime(self, name, predicate, recordValues):
    indicator = PredicateSatisfactionTime(name, predicate, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewRewardMinimumValue(self, name, reward, recordValues):
    indicator = RewardMinimumValue(name, reward, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewRewardMaximumValue(self, name, reward, recordValues):
    indicator = RewardMaximumValue(name, reward, recordValues)
    self.indicators.append(indicator)
    return indicator

  def NewRewardMeanValue(self, name, reward, recordValues):
    indicator = RewardMeanValue(name, reward, recordValues)
    self.indicators.append(indicator)
    return indicator

  def ResetIndicators(self):
    for indicator in self.indicators:
      indicator.Reset()

  def ResetIndicatorsAtBeginningOfExecution(self):
    for indicator in self.indicators:
      indicator.ResetAtBeginningOfExecution(self.currentState)

  def UpdateIndicatorsAfterTransitionFiring(self, transition, firingDate):
    for indicator in self.indicators:
      indicator.UpdateAfterTransitionFiring(self.currentState, transition, firingDate)

  def UpdateIndicatorsAtEndOfExecution(self, missionTime):
    for indicator in self.indicators:
      indicator.UpdateAtEndOfExecution(self.currentState, missionTime)

# 3. Schedules
# ------------

class Schedule:
  def __init__(self):
    self.system = None
    self.events = []
    self.scheduledEvents = []
    self.candidateEvents = []

  def GetSystem(self):
    return self.system

  def SetSystem(self, system):
    self.system = system
    self.events.clear()
    for transition in self.system.GetTransitions():
      event = Event(transition)
      self.events.append(event)
    self.Restart()

  def GetEvents(self):
    return self.events

  def Restart(self):
    self.scheduledEvents.clear()
    self.candidateEvents.clear()
    for event in self.events:
      event.Reset()

  def GetNumberOfScheduledEvents(self):
    return len(self.scheduledEvents)

  def ScheduleEvent(self, scheduledEvent, schedulingDate, firingDate):
    scheduledEvent.SetSchedulingDate(schedulingDate)
    scheduledEvent.SetFiringDate(firingDate)
    position = 0
    while position<len(self.scheduledEvents):
      currentEvent = self.scheduledEvents[position]
      if currentEvent.GetFiringDate()>scheduledEvent.GetFiringDate():
        break
      position += 1
    self.scheduledEvents.insert(position, scheduledEvent)

  def UnscheduleEvent(self, unscheduledEvent):
    self.scheduledEvents.remove(unscheduledEvent)
    unscheduledEvent.SetFiringDate(Delay.INFINITY)

  def GetCandidateEvents(self):
    self.candidateEvents.clear()
    if len(self.scheduledEvents)==0:
      return self.candidateEvents
    event = self.scheduledEvents[0]
    firstDate = event.GetFiringDate()
    self.candidateEvents.append(event)
    for i in range(1, len(self.scheduledEvents)):
      event = self.scheduledEvents[i]
      if event.GetFiringDate()>firstDate:
        break
      self.candidateEvents.append(event)
    return self.candidateEvents

# 4. Events
# ---------

class Event:
  def __init__(self, transition):
    self.transition = transition
    self.schedulingDate = Delay.INFINITY
    self.firingDate = Delay.INFINITY
    self.remainingTime = 0

  def GetTransition(self):
    return self.transition

  def Reset(self):
    self.SetSchedulingDate(Delay.INFINITY)
    self.SetFiringDate(Delay.INFINITY)
    self.remainingTime = 0

  def GetSchedulingDate(self):
    return self.schedulingDate

  def SetSchedulingDate(self, schedulingDate):
    self.schedulingDate = schedulingDate

  def GetFiringDate(self):
    return self.firingDate

  def SetFiringDate(self, firingDate):
    self.firingDate = firingDate

  def GetRemainingTime(self):
    return self.remainingTime

  def SetRemainingTime(self, time):
    self.remainingTime = time

# 5. Indicators
# -------------

# 5.1. Base Indicators
# --------------------

class Indicator:
  def __init__(self, name, recordValues):
    self.name = name
    self.numberOfValues = 0
    self.sumOfValues = 0
    self.sumOfValueSquares = 0
    self.minimumValue = sys.float_info.max
    self.maximumValue = -sys.float_info.max
    self.recordValues = recordValues
    self.recordedValues = []

  def GetName(self):
    return self.name

  def Reset(self):
    self.numberOfValues = 0
    self.sumOfValues = 0
    self.sumOfValueSquares = 0
    self.minimumValue = sys.float_info.max
    self.maximumValue = -sys.float_info.max
    self.recordedValues.clear()

  def GetRecordValues(self):
    return self.recordValues

  def SetRecordValues(self, value):
    self.recordValues = value

  def ResetAtBeginningOfExecution(self, state):
    pass

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    pass

  def UpdateAtEndOfExecution(self, state, missionTime):
    pass

  def GetNumberOfValues(self):
    return self.numberOfValues

  def GetMinimumValue(self):
    return self.minimumValue

  def GetMaximumValue(self):
    return self.maximumValue

  def GetMean(self):
    if self.numberOfValues==0:
      return 0
    return self.sumOfValues/self.numberOfValues

  def GetStandardDeviation(self):
    if self.numberOfValues==0:
      return 0
    return math.sqrt(self.sumOfValueSquares/self.numberOfValues - self.GetMean()**2)

  def GetErrorFactor90(self):
    if self.numberOfValues==0:
      return 0
    return 1.64 * self.GetStandardDeviation()/math.sqrt(self.numberOfValues)

  def GetErrorFactor95(self):
    if self.numberOfValues==0:
      return 0
    return 1.96 * self.GetStandardDeviation()/math.sqrt(self.numberOfValues)

  def GetErrorFactor99(self):
    if self.numberOfValues==0:
      return 0
    return 2.58 * self.GetStandardDeviation()/math.sqrt(self.numberOfValues)

  def GetRecordedValues(self):
    return self.recordedValues

  def RecordValue(self, value):
    self.numberOfValues += 1
    self.sumOfValues += value
    self.sumOfValueSquares += value**2
    if value<self.minimumValue:
      self.minimumValue = value
    if value>self.maximumValue:
      self.maximumValue = value
    if self.recordValues:
      self.recordedValues.append(value)

# 5.2. Firing Counts
# ------------------

class FiringCount(Indicator):
  def __init__(self, name, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.count = 0

  def ResetAtBeginningOfExecution(self, state):
    self.count = 0

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    self.count += 1

  def UpdateAtEndOfExecution(self, state, missionTime):
    self.RecordValue(self.count)

# 5.3. Transition Firing Counts
# -----------------------------

class TransitionFiringCount(Indicator):
  def __init__(self, name, transition, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.transition = transition
    self.count = 0

  def ResetAtBeginningOfExecution(self, state):
    self.count = 0

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    if transition==self.transition:
      self.count += 1

  def UpdateAtEndOfExecution(self, state, missionTime):
    self.RecordValue(self.count)

# 5.4. First Transition Firing Dates
# ----------------------------------

class TransitionFirstFiringDate(Indicator):
  def __init__(self, name, transition, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.transition = transition
    self.firstDate = -1

  def ResetAtBeginningOfExecution(self, state):
    self.firstDate = -1

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    if transition==self.transition and self.firstDate==-1:
      self.firstDate = firingDate

  def UpdateAtEndOfExecution(self, state, missionTime):
    if self.firstDate!=-1:
      self.RecordValue(self.firstDate)

# 5.5. Predicate Satisfaction Counts
# ----------------------------------

class PredicateSatisfactionCount(Indicator):
  def __init__(self, name, predicate, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.predicate = predicate
    self.count = 0
    self.reached = False

  def ResetAtBeginningOfExecution(self, state):
    self.reached = self.predicate(state)
    if self.reached:
      self.count = 1
    else:
      self.count = 0

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    if self.predicate(state):
      if not self.reached:
        self.reached = True
        self.count += 1
    else:
      self.reached = False

  def UpdateAtEndOfExecution(self, state, missionTime):
    self.RecordValue(self.count)

# 5.6. Predicate First Satisfaction Dates
# ---------------------------------------

class PredicateFirstSatisfactionDate(Indicator):
  def __init__(self, name, predicate, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.predicate = predicate
    self.firstDate = -1

  def ResetAtBeginningOfExecution(self, state):
    if self.predicate(state):
      self.firstDate = 0
    else:
      self.firstDate = -1

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    if self.firstDate!=-1:
      return
    if self.predicate(state):
      self.firstDate = firingDate

  def UpdateAtEndOfExecution(self, state, missionTime):
    if self.firstDate!=-1:
      self.RecordValue(self.firstDate)

# 5.7. Predicate Satisfaction Time
# --------------------------------

class PredicateSatisfactionTime(Indicator):
  def __init__(self, name, predicate, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.predicate = predicate
    self.lastSatisfactionDate = -1
    self.satisfactionTime = 0

  def ResetAtBeginningOfExecution(self, state):
    if self.predicate(state):
      self.lastSatisfactionDate = 0
    else:
      self.lastSatisfactionDate = -1
    self.satisfactionTime = 0

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    if self.lastSatisfactionDate==-1:
      if self.predicate(state):
        self.lastSatisfactionDate = firingDate
    else:
      if not self.predicate(state):
        self.satisfactionTime += firingDate - self.lastSatisfactionDate
        self.lastSatisfactionDate = -1

  def UpdateAtEndOfExecution(self, state, missionTime):
    if self.lastSatisfactionDate!=-1:
      self.satisfactionTime += missionTime - self.lastSatisfactionDate
    self.RecordValue(self.satisfactionTime)

# 5.8. Reward Minimum Values
# --------------------------

class RewardMinimumValue(Indicator):
  def __init__(self, name, reward, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.reward = reward
    self.minimumRewardValue = 0

  def ResetAtBeginningOfExecution(self, state):
    self.minimumRewardValue = self.reward(state)

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    currentValue = self.reward(state)
    if currentValue<self.minimumRewardValue:
      self.minimumRewardValue = currentValue

  def UpdateAtEndOfExecution(self, state, missionTime):
    self.RecordValue(self.minimumRewardValue)

# 5.9. Reward Maximum Values
# --------------------------

class RewardMaximumValue(Indicator):
  def __init__(self, name, reward, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.reward = reward
    self.maximumRewardValue = 0

  def ResetAtBeginningOfExecution(self, state):
    self.maximumRewardValue = self.reward(state)

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    currentValue = self.reward(state)
    if currentValue>self.maximumRewardValue:
      self.maximumRewardValue = currentValue

  def UpdateAtEndOfExecution(self, state, missionTime):
    self.RecordValue(self.maximumRewardValue)

# 5.10. Reward Mean Values
# ------------------------

class RewardMeanValue(Indicator):
  def __init__(self, name, reward, recordValues):
    Indicator.__init__(self, name, recordValues)
    self.reward = reward
    self.previousValue = 0
    self.previousDate = 0
    self.valueSummation = 0

  def ResetAtBeginningOfExecution(self, state):
    self.previousValue = self.reward(state)
    self.previousDate = 0
    self.valueSummation = 0

  def UpdateAfterTransitionFiring(self, state, transition, firingDate):
    self.valueSummation += self.previousValue * (firingDate - self.previousDate)
    self.previousValue = self.reward(state)
    self.previousDate = firingDate

  def UpdateAtEndOfExecution(self, state, missionTime):
    if missionTime==0:
      return 0.0
    self.valueSummation += self.previousValue * (missionTime - self.previousDate)
    self.RecordValue(self.valueSummation/missionTime)



