# TPK4186 - Advanced Tools for Performance Engineering
# AntFood.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. TerritoryState
# 3. Territory
# 4. Main

# 1. Required modules
# -------------------

import sys
import random
import numpy as np

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. TerritoryState
# -----------------

class TerritoryState(Alea.DiscreteEventSystems.State):
  
  def __init__(self, territory):
    Alea.DiscreteEventSystems.State.__init__(self, territory)
    self.antRow = 0
    self.antColumn = 0
    self.hasAntFoundFood = False
    self.currentCellExplored = False

  @staticmethod
  def New(territory):
    return TerritoryState(territory)

  def Reset(self):
    self.system.Initialize()
    self.antRow = self.system.GetSize()//2
    self.antColumn = self.system.GetSize()//2
    self.hasAntFoundFood = False
    self.currentCellExplored = False

  def CanAntExploreVainlyCurrentCell(self):
    return not self.hasAntFoundFood and not self.currentCellExplored and not self.system.DoesCellContainFood(self.antRow, self.antColumn)

  def ExploreVainlyCurrentCell(self):
    self.currentCellExplored = True
    self.system.DropPheromone(self.antRow, self.antColumn)

  def CanAntFindFoodInCurrentCell(self):
    return not self.hasAntFoundFood and not self.currentCellExplored and self.system.DoesCellContainFood(self.antRow, self.antColumn)

  def FindFoodInCurrentCell(self):
    self.hasAntFoundFood = True
    self.currentCellExplored = True

  def CanAntGoNorth(self):
    return not self.hasAntFoundFood and self.currentCellExplored and self.system.IsCellInTerritory(self.antRow-1, self.antColumn)

  def MoveAntNorth(self):
    self.antRow -= 1
    self.currentCellExplored = False

  def CanAntGoSouth(self):
    return not self.hasAntFoundFood and self.currentCellExplored and self.system.IsCellInTerritory(self.antRow+1, self.antColumn)

  def MoveAntSouth(self):
    self.antRow += 1
    self.currentCellExplored = False

  def CanAntGoWest(self):
    return not self.hasAntFoundFood and self.currentCellExplored and self.system.IsCellInTerritory(self.antRow, self.antColumn-1)

  def MoveAntWest(self):
    self.antColumn -= 1
    self.currentCellExplored = False

  def CanAntGoEast(self):
    return not self.hasAntFoundFood and self.currentCellExplored and self.system.IsCellInTerritory(self.antRow, self.antColumn+1)

  def MoveAntEast(self):
    self.antColumn += 1
    self.currentCellExplored = False

# 3. Territory
# ------------

class NorthCellWeight(Alea.DiscreteEventSystems.Weight):
  def Calculate(self, state):
    if state.system.DoesCellContainPheromone(state.antRow-1, state.antColumn):
      return 1
    return 1000

class SouthCellWeight(Alea.DiscreteEventSystems.Weight):
  def Calculate(self, state):
    if state.system.DoesCellContainPheromone(state.antRow+1, state.antColumn):
      return 1
    return 1000

class WestCellWeight(Alea.DiscreteEventSystems.Weight):
  def Calculate(self, state):
    if state.system.DoesCellContainPheromone(state.antRow, state.antColumn-1):
      return 1
    return 1000
  
class EastCellWeight(Alea.DiscreteEventSystems.Weight):
  def Calculate(self, state):
    if state.system.DoesCellContainPheromone(state.antRow, state.antColumn+1):
      return 1
    return 1000
 

class Territory(Alea.DiscreteEventSystems.DiscreteEventSystem):
  def __init__(self, size, numberOfFoodChunks):
    Alea.DiscreteEventSystems.DiscreteEventSystem.__init__(self, TerritoryState.New)
    self.size = size
    self.cells = np.zeros((size, size))
    self.numberOfFoodChunks = numberOfFoodChunks
    self.NewTransition("ExploreVainlyCell", TerritoryState.CanAntExploreVainlyCurrentCell, TerritoryState.ExploreVainlyCurrentCell, Alea.DiscreteEventSystems.DiracDelay(1))
    self.NewTransition("FindFood", TerritoryState.CanAntFindFoodInCurrentCell, TerritoryState.FindFoodInCurrentCell, Alea.DiscreteEventSystems.DiracDelay(1))
    self.NewTransition("GoNorth", TerritoryState.CanAntGoNorth, TerritoryState.MoveAntNorth, Alea.DiscreteEventSystems.DiracDelay(1), NorthCellWeight())
    self.NewTransition("GoSouth", TerritoryState.CanAntGoSouth, TerritoryState.MoveAntSouth, Alea.DiscreteEventSystems.DiracDelay(1), SouthCellWeight())
    self.NewTransition("GoWest", TerritoryState.CanAntGoWest, TerritoryState.MoveAntWest, Alea.DiscreteEventSystems.DiracDelay(1), WestCellWeight())
    self.NewTransition("GoEast", TerritoryState.CanAntGoEast, TerritoryState.MoveAntEast, Alea.DiscreteEventSystems.DiracDelay(1), EastCellWeight())

  def Initialize(self):
    self.cells.fill(0)
    for n in range(0, self.numberOfFoodChunks):
      row = random.randrange(0, self.size)
      column = random.randrange(0, self.size)
      self.cells[row, column] += 1

  def GetSize(self):
    return self.size

  def IsCellInTerritory(self, row, column):
    return row>=0 and row<self.size and column>=0 and column<self.size

  def DoesCellContainFood(self, row, column):
    return self.cells[row, column] > 0

  def DropPheromone(self, row, column):
    self.cells[row, column] = -1

  def DoesCellContainPheromone(self, row, column):
    return self.cells[row, column]==-1

# 4. Main
# -------

gridSize = 10
numberOfFoodChunks = 5
territory = Territory(gridSize, numberOfFoodChunks)
simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(territory)
FF = territory.LookForTransition("FindFood")
I1 = simulator.NewTransitionFiringCount("I1", FF, False)
I2 = simulator.NewTransitionFirstFiringDate("I2", FF, False)

simulator.MonteCarloSimulation(100, 10000)

print("I1\t{0:.2f}".format(I1.GetMean()))
print("I2\t{0:.2f}".format(I2.GetMean()))
