# TPK4186 - Advanced Tools for Performance Engineering
# DieselGenerators.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. DieselGroupState
# 3. DieselGroup
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. DieselGroupState
# -------------------

class DieselGroupState(Alea.DiscreteEventSystems.State):
  FAILED = 0
  WORKING = 1
  INITIAL_OPERATION = 0
  REGULAR_OPERATION = 1
  MAINTENANCE = 2

  def __init__(self, dieselGroup):
    Alea.DiscreteEventSystems.State.__init__(self, dieselGroup)
    self.generator1State = DieselGroupState.WORKING
    self.generator1Mode = DieselGroupState.INITIAL_OPERATION
    self.generator2State = DieselGroupState.WORKING
    self.generator2Mode = DieselGroupState.INITIAL_OPERATION

  @staticmethod
  def New(dieselGroup):
    return DieselGroupState(dieselGroup)

  def Reset(self):
    self.generator1State = DieselGroupState.WORKING
    self.generator1Mode = DieselGroupState.INITIAL_OPERATION
    self.generator2State = DieselGroupState.WORKING
    self.generator2Mode = DieselGroupState.INITIAL_OPERATION


  def IsGenerator1InOperation(self):
    return self.generator1Mode==DieselGroupState.INITIAL_OPERATION or self.generator1Mode==DieselGroupState.REGULAR_OPERATION

  def CanGenerator1Fail(self):
    return self.IsGenerator1InOperation() and self.generator1State==DieselGroupState.WORKING

  def FailGenerator1(self):
    self.generator1State = DieselGroupState.FAILED

  def CanInitialMaintenanceGenerator1BeStarted(self):
    return self.generator1Mode==DieselGroupState.INITIAL_OPERATION

  def CanRegularMaintenanceGenerator1BeStarted(self):
    return self.generator1Mode==DieselGroupState.REGULAR_OPERATION

  def StartMaintenanceGenerator1(self):
    self.generator1Mode = DieselGroupState.MAINTENANCE

  def CanMaintenanceGenerator1BeCompleted(self):
    return self.generator1Mode==DieselGroupState.MAINTENANCE and self.generator1State==DieselGroupState.WORKING

  def CompleteMaintenanceGenerator1(self):
    self.generator1Mode = DieselGroupState.REGULAR_OPERATION

  def CanRepairGenerator1BeCompleted(self):
    return self.generator1Mode==DieselGroupState.MAINTENANCE and self.generator1State==DieselGroupState.FAILED

  def CompleteRepairGenerator1(self):
    self.generator1State = DieselGroupState.WORKING
    self.generator1Mode = DieselGroupState.REGULAR_OPERATION


  def IsGenerator2InOperation(self):
    return self.generator2Mode==DieselGroupState.INITIAL_OPERATION or self.generator2Mode==DieselGroupState.REGULAR_OPERATION

  def CanGenerator2Fail(self):
    return self.IsGenerator2InOperation() and self.generator2State==DieselGroupState.WORKING

  def FailGenerator2(self):
    self.generator2State = DieselGroupState.FAILED

  def CanInitialMaintenanceGenerator2BeStarted(self):
    return self.generator2Mode==DieselGroupState.INITIAL_OPERATION

  def CanRegularMaintenanceGenerator2BeStarted(self):
    return self.generator2Mode==DieselGroupState.REGULAR_OPERATION

  def StartMaintenanceGenerator2(self):
    self.generator2Mode = DieselGroupState.MAINTENANCE

  def CanMaintenanceGenerator2BeCompleted(self):
    return self.generator2Mode==DieselGroupState.MAINTENANCE and self.generator2State==DieselGroupState.WORKING

  def CompleteMaintenanceGenerator2(self):
    self.generator2Mode = DieselGroupState.REGULAR_OPERATION

  def CanRepairGenerator2BeCompleted(self):
    return self.generator2Mode==DieselGroupState.MAINTENANCE and self.generator2State==DieselGroupState.FAILED

  def CompleteRepairGenerator2(self):
    self.generator2State = DieselGroupState.WORKING
    self.generator2Mode = DieselGroupState.REGULAR_OPERATION


  def IsGenerator1OutOfService(self):
    return self.generator1State==DieselGroupState.FAILED or self.generator1Mode==DieselGroupState.MAINTENANCE

  def IsGenerator2OutOfService(self):
    return self.generator2State==DieselGroupState.FAILED or self.generator2Mode==DieselGroupState.MAINTENANCE

  def IsOutOfService(self):
    return self.IsGenerator1OutOfService() and self.IsGenerator2OutOfService()

  def IsInMaintenance(self):
    return self.generator1Mode==DieselGroupState.MAINTENANCE and self.generator2Mode==DieselGroupState.MAINTENANCE

# 3. DieselGenerator
# ----------------

meanFailureTime = 20000
timeToInitialMaintenanceGenerator1 = 730
timeToInitialMaintenanceGenerator2 = 1460
timeToRegularMaintenance = 2190
maintenanceDuration = 4
repairDuration = 48

dieselGroup = Alea.DiscreteEventSystems.DiscreteEventSystem(DieselGroupState.New)

dieselGroup.NewTransition("FailureGenerator1", DieselGroupState.CanGenerator1Fail, DieselGroupState.FailGenerator1, Alea.DiscreteEventSystems.ExponentialDelay(meanFailureTime))
dieselGroup.NewTransition("StartInitialMaintenanceGenerator1", DieselGroupState.CanInitialMaintenanceGenerator1BeStarted, DieselGroupState.StartMaintenanceGenerator1, Alea.DiscreteEventSystems.DiracDelay(timeToInitialMaintenanceGenerator1))
dieselGroup.NewTransition("StartRegularMaintenanceGenerator1", DieselGroupState.CanRegularMaintenanceGenerator1BeStarted, DieselGroupState.StartMaintenanceGenerator1, Alea.DiscreteEventSystems.DiracDelay(timeToRegularMaintenance))
dieselGroup.NewTransition("CompleteMaintenanceGenerator1", DieselGroupState.CanMaintenanceGenerator1BeCompleted, DieselGroupState.CompleteMaintenanceGenerator1, Alea.DiscreteEventSystems.DiracDelay(maintenanceDuration))
dieselGroup.NewTransition("CompleteRepairGenerator1", DieselGroupState.CanRepairGenerator1BeCompleted, DieselGroupState.CompleteRepairGenerator1, Alea.DiscreteEventSystems.DiracDelay(repairDuration))

dieselGroup.NewTransition("FailureGenerator2", DieselGroupState.CanGenerator2Fail, DieselGroupState.FailGenerator2, Alea.DiscreteEventSystems.ExponentialDelay(meanFailureTime))
dieselGroup.NewTransition("StartInitialMaintenanceGenerator2", DieselGroupState.CanInitialMaintenanceGenerator2BeStarted, DieselGroupState.StartMaintenanceGenerator2, Alea.DiscreteEventSystems.DiracDelay(timeToInitialMaintenanceGenerator2))
dieselGroup.NewTransition("StartRegularMaintenanceGenerator2", DieselGroupState.CanRegularMaintenanceGenerator2BeStarted, DieselGroupState.StartMaintenanceGenerator2, Alea.DiscreteEventSystems.DiracDelay(timeToRegularMaintenance))
dieselGroup.NewTransition("CompleteMaintenanceGenerator2", DieselGroupState.CanMaintenanceGenerator2BeCompleted, DieselGroupState.CompleteMaintenanceGenerator2, Alea.DiscreteEventSystems.DiracDelay(maintenanceDuration))
dieselGroup.NewTransition("CompleteRepairGenerator2", DieselGroupState.CanRepairGenerator2BeCompleted, DieselGroupState.CompleteRepairGenerator2, Alea.DiscreteEventSystems.DiracDelay(repairDuration))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(dieselGroup)
I1 = simulator.NewPredicateSatisfactionTime("I1", DieselGroupState.IsOutOfService, False)
I2 = simulator.NewPredicateSatisfactionTime("I2", DieselGroupState.IsInMaintenance, False)

missionTime = 50000
numberOfTrials = 10000
simulator.MonteCarloSimulation(missionTime, numberOfTrials)

print("I1\t{0:.3e}".format(I1.GetMean()))
print("I2\t{0:.3e}".format(I2.GetMean()))
