# TPK4186 - Advanced Tools for Performance Engineering
# ForestFire.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Maps
# 3. Simulators
# 4. Animators
# 5. Main

# To get this program work, you need run the following command in the Spyder console
# %matplotlib qt

# 1. Required modules
# -------------------

import random
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors
import matplotlib.animation as animation

# 2. Maps
# -------

class Map:
  EMPTY = 0.1
  FOREST = 0.3
  BURNING = 0.5
  BURNT = 0.7
  CANDIDATE = 0.9

  def __init__(self, numberOfRows, numberOfColumns):
    self.numberOfRows = numberOfRows 
    self.numberOfColumns = numberOfColumns
    self.forestProbability = 0.60
    self.ignitionProbability = 0.80
    self.extinctionProbability = 0.10
    self.grid = np.zeros((numberOfRows, numberOfColumns))
    self.burningParcels = []
    self.numberOfBurntParcels = 0
    self.numberOfStepsBeforeExtinction = 0

  def GetNumberOfBurningParcels(self):
    return len(self.burningParcels)

  def GetNumberOfBurntParcels(self):
    return self.numberOfBurntParcels

  def GetNumberOfStepsBeforeExtinction(self):
    return self.numberOfStepsBeforeExtinction

  def Initialize(self):
    for r in range(0, self.numberOfRows):
      for c in range(0, self.numberOfColumns):
        z = random.random()
        if z<self.forestProbability:
          self.grid[r][c] = Map.FOREST
        else:
          self.grid[r][c] = Map.EMPTY
    self.numberOfBurntParcels = 0
    self.numberOfStepsBeforeExtinction = 0

  def IgniteFire(self):
    r = random.randrange(self.numberOfRows)
    c = random.randrange(self.numberOfColumns)
    self.grid[r][c] = Map.BURNING
    self.burningParcels.append((r, c))

  def UpdateState(self):
    if len(self.burningParcels)==0:
        return
    self.numberOfStepsBeforeExtinction += 1
    candidateParcels = []
    for (r, c) in self.burningParcels:
      if r>0 and self.grid[r-1][c]==Map.FOREST:
        self.grid[r][c]==Map.CANDIDATE
        candidateParcels.append((r-1, c))
      if c>0 and self.grid[r][c-1]==Map.FOREST:
        self.grid[r][c]==Map.CANDIDATE
        candidateParcels.append((r, c-1))
      if r<self.numberOfRows-1 and self.grid[r+1][c]==Map.FOREST:
        self.grid[r][c]==Map.CANDIDATE
        candidateParcels.append((r+1, c))
      if c<self.numberOfColumns-1 and self.grid[r][c+1]==Map.FOREST:
        self.grid[r][c]==Map.CANDIDATE
        candidateParcels.append((r, c+1))
    burningParcels = self.burningParcels
    self.burningParcels = []
    for (r, c) in burningParcels:
      z = random.random()
      if z<self.extinctionProbability:
        self.grid[r][c] = Map.BURNT
        self.numberOfBurntParcels += 1
      else:
        self.burningParcels.append((r, c))
    for (r, c) in candidateParcels:
      z = random.random()
      if z<self.ignitionProbability:
        self.grid[r][c] = Map.BURNING
        self.burningParcels.append((r, c))
      else:
        self.grid[r][c] = Map.FOREST

# 3. Simulators
# --------------

class Simulator:
  def __init__(self, map):
    self.map = map
    self.numbersOfStepBeforeExtinctions = None
    self.numbersOfBurntParcels = None

  def DrawHistory(self):
    self.map.Initialize()
    self.map.IgniteFire()
    while self.map.GetNumberOfBurningParcels()>0:
      self.map.UpdateState()

  def Simulate(self, numberOfTrials):
    self.numbersOfStepBeforeExtinctions = np.zeros(numberOfTrials, dtype=int)
    self.numbersOfBurntParcels = np.zeros(numberOfTrials, dtype=int)
    for trial in range(0, numberOfTrials):
      self.DrawHistory()
      self.numbersOfStepBeforeExtinctions[trial] = self.map.GetNumberOfStepsBeforeExtinction()
      self.numbersOfBurntParcels[trial] = self.map.GetNumberOfBurntParcels()

# 4. Animators
# ------------

class Animator:
  TheAnimator = None

  def __init__(self, map, width, height):
    self.map = map
    colorMap = colors.ListedColormap(['white', 'green', 'red', 'maroon', 'blue'])
    bounds = [0, 0.2, 0.4, 0.6, 0.8, 1.0]
    norm = colors.BoundaryNorm(bounds, colorMap.N)
    self.figure = plt.figure(figsize=(width, height))
    self.image = plt.imshow(self.map.grid, cmap=colorMap, norm=norm, aspect='equal')
    self.animation = None
    Animator.TheAnimator = self

  def ShowCurrentMap(self):
    self.image.set_array(self.map.grid)
    plt.show()

  @staticmethod
  def ShowNextMap(i):
    self = Animator.TheAnimator
    self.map.UpdateState()
    self.image.set_array(self.map.grid)
    # plt.savefig("step{0:d}.pdf".format(self.map.GetNumberOfStepsBeforeExtinction()))
    return self.image,

  @staticmethod
  def Iterator():
    self = Animator.TheAnimator
    numberOfSteps = 0
    while self.map.GetNumberOfBurningParcels()>0:
      numberOfSteps += 1
      # print("{0:d}\t{1:d}".format(numberOfSteps, self.map.GetNumberOfBurningParcels()))
      yield numberOfSteps

  def LaunchAnimation(self, timeStep):
    self.image.set_array(self.map.grid)
    # plt.savefig("step0.pdf")
    self.animation = animation.FuncAnimation(self.figure, Animator.ShowNextMap, frames = Animator.Iterator, interval = timeStep, repeat = False)
    plt.show()

# 5. Main
# -------

map = Map(10, 10)
map.forestProbability = 0.55
map.ignitionProbability = 0.70
map.extinctionProbability = 0.20

animator = Animator(map, 10, 10)
map.Initialize()
map.IgniteFire()
# animator.ShowInitialMap()
animator.LaunchAnimation(300)

# simulator = Simulator(map)
# numberOfTrials = 1000
# simulator.Simulate(numberOfTrials)
# simulator.numbersOfBurntParcels.sort()
# plt.plot(np.arange(numberOfTrials)*100/numberOfTrials, simulator.numbersOfBurntParcels, color='blue')
# plt.xlabel('Percentage')
# plt.ylabel('Number of burnt parcels')
# plt.title('Distribution of the number of burnt parcels')
# plt.show()

