# TPK4186 - Advanced Tools for Performance Engineering
# FourEngineStateEngine.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. EngineStates
# 3. Engine
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. EngineStates
# ---------

class EngineState(Alea.DiscreteEventSystems.State):
  OVERLAP = 0
  INTAKE = 1
  COMPRESSION = 2
  COMBUSTION = 3
  EXHAUST = 4

  CLOSE = 0
  OPEN = 1

  FIRST_UP = 0
  DOWN = 1
  UP = 2

  def __init__(self, engine):
    Alea.DiscreteEventSystems.State.__init__(self, engine)
    self.phase = EngineState.OVERLAP
    self.inletValveState = EngineState.OPEN
    self.exhaustValveState = EngineState.OPEN
    self.pistonMovement = EngineState.FIRST_UP

  @staticmethod
  def New(engine):
    return EngineState(engine)

  def Reset(self):
    self.phase = EngineState.OVERLAP
    self.inletValveState = EngineState.OPEN
    self.exhaustValveState = EngineState.OPEN
    self.pistonMovement = EngineState.FIRST_UP

  # Intake
  # ------

  def CanIntakeStart(self):
    return self.phase==EngineState.OVERLAP

  def StartIntake(self):
    self.phase = EngineState.INTAKE
    self.exhaustValveState = EngineState.CLOSE

  # Compression
  # -----------

  def CanCompressionStart(self):
    return self.phase==EngineState.INTAKE

  def StartCompression(self):
    self.phase = EngineState.COMPRESSION
    self.inletValveState = EngineState.CLOSE

  # Combustion
  # ----------

  def CanCombustionStart(self):
    return self.phase==EngineState.COMPRESSION

  def StartCombustion(self):
    self.phase = EngineState.COMBUSTION

  # Exhaust
  # -------

  def CanExhaustStart(self):
    return self.phase==EngineState.COMBUSTION

  def StartExhaust(self):
    self.phase = EngineState.EXHAUST
    self.exhaustValveState = EngineState.OPEN

  # Overlap
  # -------

  def CanOverlapStart(self):
    return self.phase==EngineState.EXHAUST

  def StartOverlap(self):
    self.phase = EngineState.OVERLAP
    self.inletValveState = EngineState.OPEN

  # Predicates
  # ----------

  def IsInletValveOpen(self):
    return self.inletValveState==EngineState.OPEN

  def IsExhaustValveOpen(self):
    return self.exhaustValveState==EngineState.OPEN

  # Piston
  # ------

  def IsPistonFirstMovingUp(self):
    return self.pistonMovement==EngineState.FIRST_UP

  def FirstSwitchPistonDirection(self):
    self.pistonMovement = EngineState.DOWN

  def IsPistonMoving(self):
    return self.pistonMovement!=EngineState.FIRST_UP

  def SwitchPistonDirection(self):
    if self.pistonMovement==EngineState.UP:
      self.pistonMovement = EngineState.DOWN
    else:
      self.pistonMovement = EngineState.UP

# 3. Engine
# ---------

engine = Alea.DiscreteEventSystems.DiscreteEventSystem(EngineState.New)
T1 = engine.NewTransition("StartIntake", EngineState.CanIntakeStart, EngineState.StartIntake, Alea.DiscreteEventSystems.DiracDelay(35))
T2 = engine.NewTransition("StartCompression", EngineState.CanCompressionStart, EngineState.StartCompression, Alea.DiscreteEventSystems.DiracDelay(190))
T3 = engine.NewTransition("StartCombustion", EngineState.CanCombustionStart, EngineState.StartCombustion, Alea.DiscreteEventSystems.DiracDelay(150))
T4 = engine.NewTransition("StartExhaust", EngineState.CanExhaustStart, EngineState.StartExhaust, Alea.DiscreteEventSystems.DiracDelay(130))
T5 = engine.NewTransition("StartOverlap", EngineState.CanOverlapStart, EngineState.StartOverlap, Alea.DiscreteEventSystems.DiracDelay(215))
T6 = engine.NewTransition("FirstSwitchPistonDirection", EngineState.IsPistonFirstMovingUp, EngineState.FirstSwitchPistonDirection, Alea.DiscreteEventSystems.DiracDelay(15)) 
T7 = engine.NewTransition("SwitchPistonDirection", EngineState.IsPistonMoving, EngineState.SwitchPistonDirection, Alea.DiscreteEventSystems.DiracDelay(180)) 

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(engine)
# I1 = simulator.NewTransitionFiringCount("I1", T1, False)
# I2 = simulator.NewTransitionFiringCount("I2", T3, False)
# I3 = simulator.NewRewardMaximumValue("I3", Queue.GetNumberOfClientsWaiting, False)
# I4 = simulator.NewRewardMeanValue("I4", Queue.GetNumberOfClientsWaiting, False)
# I5 = simulator.NewRewardMeanValue("I5", Queue.IsCounterIdle, False)

simulator.MonteCarloSimulation(1500, 1)

# print("I1\t{0:.2f}".format(I1.GetMean()))
# print("I2\t{0:.2f}".format(I2.GetMean()))
# print("I3\t{0:.2f}".format(I3.GetMaximumValue()))
# print("I4\t{0:.2f}".format(I4.GetMean()))
# print("I5\t{0:.2f}".format(I5.GetMean()))

