# TPK4186 - Advanced Tools for Performance Engineering
# GameOfLife.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Maps
# 3. Animators
# 4. Main

# To get this program work, you need run the following command in the Spyder console
# %matplotlib qt

# 1. Required modules
# -------------------

import random
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors
import matplotlib.animation as animation

# 2. Maps
# -------

class Map:
  DEAD = 0.2
  STAYING_DEAD = 0.3
  BECOMING_DEAD = 0.4
  ALIVE = 0.6
  STAYING_ALIVE = 0.7
  BECOMING_ALIVE = 0.8

  def __init__(self, numberOfRows, numberOfColumns):
    self.numberOfRows = numberOfRows 
    self.numberOfColumns = numberOfColumns
    self.grid = np.zeros((numberOfRows, numberOfColumns))
    self.livingCells = []

  def GetNumberOfLivingCells(self):
    return len(self.livingCells)

  def Initialize(self):
    for row in range(0, self.numberOfRows):
      for column in range(0, self.numberOfColumns):
        self.grid[row][column] = Map.DEAD

  def SetCell(self, row, column, state):
    self.grid[row][column] = state
    if state==Map.ALIVE:
      self.livingCells.append((row, column))

  def GetNeighborhoodOfCell(self, row, column):
    neighborhood = []
    if row-1>=0:
      if column-1>=0:
        neighborhood.append((row-1, column-1))
      neighborhood.append((row-1, column))
      if column+1<self.numberOfColumns:
        neighborhood.append((row-1, column+1))
    if column-1>=0:
      neighborhood.append((row, column-1))
    if column+1<self.numberOfColumns:
      neighborhood.append((row, column+1))
    if row+1<self.numberOfRows:
      if column-1>=0:
        neighborhood.append((row+1, column-1))
      neighborhood.append((row+1, column))
      if column+1<self.numberOfColumns:
        neighborhood.append((row+1, column+1))
    return neighborhood

  def GetNextStateOfCell(self, row, column):
    neighborhood = self.GetNeighborhoodOfCell(row, column)
    numberOfNeighborsAlive = 0
    for (r, c) in neighborhood:
      if self.grid[r][c]==Map.ALIVE or self.grid[r][c]==Map.STAYING_ALIVE or self.grid[r][c]==Map.BECOMING_DEAD:
        numberOfNeighborsAlive += 1
    if self.grid[row][column]==Map.ALIVE:
      if numberOfNeighborsAlive<2 or numberOfNeighborsAlive>3:
        nextState = Map.BECOMING_DEAD
      else:
        nextState = Map.STAYING_ALIVE
    else:
      if numberOfNeighborsAlive==3:
        nextState = Map.BECOMING_ALIVE
      else:
        nextState = Map.STAYING_DEAD
    return nextState

  def UpdateState(self):
    if len(self.livingCells)==0:
        return
    impactedCells = []
    for (row, column) in self.livingCells:
      neighborhood = self.GetNeighborhoodOfCell(row, column)
      for (r, c) in neighborhood:
        if self.grid[r][c]!=Map.DEAD and self.grid[r][c]!=Map.ALIVE:
          continue
        self.grid[r][c] = self.GetNextStateOfCell(r, c)
        impactedCells.append((r, c))
    self.livingCells.clear()
    for (r, c) in impactedCells:
      if self.grid[r][c]==Map.STAYING_DEAD or self.grid[r][c]==Map.BECOMING_DEAD:
        self.SetCell(r, c, Map.DEAD)
      else:
        self.SetCell(r, c, Map.ALIVE)

# 3. Animators
# ------------

class Animator:
  TheAnimator = None

  def __init__(self, map, width, height):
    self.map = map
    colorMap = colors.ListedColormap(['white', 'black'])
    bounds = [0, 0.5, 1.0]
    norm = colors.BoundaryNorm(bounds, colorMap.N)
    self.figure = plt.figure(figsize=(width, height))
    self.image = plt.imshow(self.map.grid, cmap=colorMap, norm=norm, aspect='equal')
    self.animation = None
    self.maximumNumberOfSteps = 0
    Animator.TheAnimator = self

  def ShowCurrentMap(self):
    self.image.set_array(self.map.grid)
    # plt.savefig("Pulsar0.pdf")
    plt.show()

  @staticmethod
  def ShowNextMap(i):
    self = Animator.TheAnimator
    self.map.UpdateState()
    self.image.set_array(self.map.grid)
    # plt.savefig("step{0:d}.pdf".format(self.map.GetNumberOfStepsBeforeExtinction()))
    return self.image,

  @staticmethod
  def Iterator():
    self = Animator.TheAnimator
    numberOfSteps = 0
    while self.map.GetNumberOfLivingCells()>0 and numberOfSteps<self.maximumNumberOfSteps:
      numberOfSteps += 1
      # print("{0:d}\t{1:d}".format(numberOfSteps, self.map.GetNumberOfLivingCells()))
      yield numberOfSteps

  def LaunchAnimation(self, maximumNumberOfSteps, timeStep):
    self.maximumNumberOfSteps = maximumNumberOfSteps
    self.image.set_array(self.map.grid)
    # plt.savefig("step0.pdf")
    self.animation = animation.FuncAnimation(self.figure, Animator.ShowNextMap, frames = Animator.Iterator, interval = timeStep, repeat = False)
    plt.show()

# 5. Main
# -------

# Pulsar
map = Map(17, 18)
map.Initialize()

map.SetCell(1, 6, Map.ALIVE)
map.SetCell(1, 12, Map.ALIVE)

map.SetCell(2, 6, Map.ALIVE)
map.SetCell(2, 12, Map.ALIVE)

map.SetCell(3, 6, Map.ALIVE)
map.SetCell(3, 7, Map.ALIVE)
map.SetCell(3, 11, Map.ALIVE)
map.SetCell(3, 12, Map.ALIVE)

map.SetCell(5, 2, Map.ALIVE)
map.SetCell(5, 3, Map.ALIVE)
map.SetCell(5, 4, Map.ALIVE)
map.SetCell(5, 7, Map.ALIVE)
map.SetCell(5, 8, Map.ALIVE)
map.SetCell(5, 10, Map.ALIVE)
map.SetCell(5, 11, Map.ALIVE)
map.SetCell(5, 14, Map.ALIVE)
map.SetCell(5, 15, Map.ALIVE)
map.SetCell(5, 16, Map.ALIVE)

map.SetCell(6, 4, Map.ALIVE)
map.SetCell(6, 6, Map.ALIVE)
map.SetCell(6, 8, Map.ALIVE)
map.SetCell(6, 10, Map.ALIVE)
map.SetCell(6, 12, Map.ALIVE)
map.SetCell(6, 14, Map.ALIVE)

map.SetCell(7, 6, Map.ALIVE)
map.SetCell(7, 7, Map.ALIVE)
map.SetCell(7, 11, Map.ALIVE)
map.SetCell(7, 12, Map.ALIVE)

map.SetCell(9, 6, Map.ALIVE)
map.SetCell(9, 7, Map.ALIVE)
map.SetCell(9, 11, Map.ALIVE)
map.SetCell(9, 12, Map.ALIVE)

map.SetCell(10, 4, Map.ALIVE)
map.SetCell(10, 6, Map.ALIVE)
map.SetCell(10, 8, Map.ALIVE)
map.SetCell(10, 10, Map.ALIVE)
map.SetCell(10, 12, Map.ALIVE)
map.SetCell(10, 14, Map.ALIVE)

map.SetCell(11, 2, Map.ALIVE)
map.SetCell(11, 3, Map.ALIVE)
map.SetCell(11, 4, Map.ALIVE)
map.SetCell(11, 7, Map.ALIVE)
map.SetCell(11, 8, Map.ALIVE)
map.SetCell(11, 10, Map.ALIVE)
map.SetCell(11, 11, Map.ALIVE)
map.SetCell(11, 14, Map.ALIVE)
map.SetCell(11, 15, Map.ALIVE)
map.SetCell(11, 16, Map.ALIVE)

map.SetCell(13, 6, Map.ALIVE)
map.SetCell(13, 7, Map.ALIVE)
map.SetCell(13, 11, Map.ALIVE)
map.SetCell(13, 12, Map.ALIVE)

map.SetCell(14, 6, Map.ALIVE)
map.SetCell(14, 12, Map.ALIVE)

map.SetCell(15, 6, Map.ALIVE)
map.SetCell(15, 12, Map.ALIVE)

animator = Animator(map, 10, 10)
#map.UpdateState()
#map.UpdateState()
#animator.ShowCurrentMap()
animator.LaunchAnimation(200, 400)


