# TPK4186 - Advanced Tools for Performance Engineering
# StandbyReliability.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. MotorGroupState
# 3. MotorGroup
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. MotorGroupState
# ---------------------

class MotorGroupState(Alea.DiscreteEventSystems.State):
  STANDBY = 0
  WORKING = 1
  FAILED = 2

  def __init__(self, motorGroup):
    Alea.DiscreteEventSystems.State.__init__(self, motorGroup)
    self.motor1State = MotorGroupState.STANDBY
    self.motor2State = MotorGroupState.STANDBY

  @staticmethod
  def New(motorGroup):
    return MotorGroupState(motorGroup)

  def Reset(self):
    self.motor1State = MotorGroupState.STANDBY
    self.motor2State = MotorGroupState.STANDBY

  def CanMotor1BeStarted(self):
    return self.motor1State==MotorGroupState.STANDBY

  def StartMotor1(self):
    self.motor1State = MotorGroupState.WORKING

  def CanMotor1FailOnDemand(self):
    return self.motor1State==MotorGroupState.STANDBY

  def CanMotor1Fail(self):
    return self.motor1State==MotorGroupState.WORKING

  def FailMotor1(self):
    self.motor1State = MotorGroupState.FAILED

  def CanMotor1BeRepaired(self):
    return self.motor1State==MotorGroupState.FAILED

  def RepairMotor1(self):
    self.motor1State = MotorGroupState.STANDBY


  def CanMotor2BeStarted(self):
    return self.motor2State==MotorGroupState.STANDBY and self.motor1State==MotorGroupState.FAILED

  def StartMotor2(self):
    self.motor2State = MotorGroupState.WORKING

  def CanMotor2FailOnDemand(self):
    return self.motor2State==MotorGroupState.STANDBY and self.motor1State==MotorGroupState.FAILED

  def CanMotor2BeStopped(self):
    return self.motor2State==MotorGroupState.WORKING and self.motor1State==MotorGroupState.WORKING

  def StopMotor2(self):
    self.motor2State = MotorGroupState.STANDBY

  def CanMotor2Fail(self):
    return self.motor2State==MotorGroupState.WORKING

  def FailMotor2(self):
    self.motor2State = MotorGroupState.FAILED

  def CanMotor2BeRepaired(self):
    return self.motor2State==MotorGroupState.FAILED

  def RepairMotor2(self):
    self.motor2State = MotorGroupState.STANDBY


  def IsMotor1Working(self):
    return self.motor1State==MotorGroupState.WORKING

  def IsMotor2Working(self):
    return self.motor2State==MotorGroupState.WORKING

  def IsOutOfService(self):
    return self.motor1State==MotorGroupState.FAILED and self.motor2State==MotorGroupState.FAILED

  def CalculateReward(self):
    if self.motor1State==MotorGroupState.WORKING:
      return 1.0
    if self.motor2State==MotorGroupState.WORKING:
      return 1.0
    if self.IsOutOfService():
      return 1.0
    return 0.0

# 3. MotorGroup
# ----------------

meanFailureTime = 8000
probabilityOfFailureOnDemand = 0.02
minimumRepairDuration = 24
maximumRepairDuration = 48

motorGroup = Alea.DiscreteEventSystems.DiscreteEventSystem(MotorGroupState.New)

motorGroup.NewTransition("StartMotor1", MotorGroupState.CanMotor1BeStarted, MotorGroupState.StartMotor1, Alea.DiscreteEventSystems.DiracDelay(0), Alea.DiscreteEventSystems.ConstantWeight(1.0-probabilityOfFailureOnDemand))
motorGroup.NewTransition("FailMotor1OnDemand", MotorGroupState.CanMotor1FailOnDemand, MotorGroupState.FailMotor1, Alea.DiscreteEventSystems.DiracDelay(0), Alea.DiscreteEventSystems.ConstantWeight(probabilityOfFailureOnDemand))
motorGroup.NewTransition("FailMotor1", MotorGroupState.CanMotor1Fail, MotorGroupState.FailMotor1, Alea.DiscreteEventSystems.ExponentialDelay(meanFailureTime))
motorGroup.NewTransition("RepairMotor1", MotorGroupState.CanMotor1BeRepaired, MotorGroupState.RepairMotor1, Alea.DiscreteEventSystems.UniformDelay(minimumRepairDuration, maximumRepairDuration))

motorGroup.NewTransition("StartMotor2", MotorGroupState.CanMotor2BeStarted, MotorGroupState.StartMotor2, Alea.DiscreteEventSystems.DiracDelay(0), Alea.DiscreteEventSystems.ConstantWeight(1.0-probabilityOfFailureOnDemand))
motorGroup.NewTransition("FailMotor2OnDemand", MotorGroupState.CanMotor2FailOnDemand, MotorGroupState.FailMotor2, Alea.DiscreteEventSystems.DiracDelay(0), Alea.DiscreteEventSystems.ConstantWeight(probabilityOfFailureOnDemand))
motorGroup.NewTransition("StopMotor2", MotorGroupState.CanMotor2BeStopped, MotorGroupState.StopMotor2, Alea.DiscreteEventSystems.DiracDelay(0))
motorGroup.NewTransition("FailMotor2", MotorGroupState.CanMotor2Fail, MotorGroupState.FailMotor2, Alea.DiscreteEventSystems.ExponentialDelay(meanFailureTime))
motorGroup.NewTransition("RepairMotor2", MotorGroupState.CanMotor2BeRepaired, MotorGroupState.RepairMotor2, Alea.DiscreteEventSystems.UniformDelay(minimumRepairDuration, maximumRepairDuration))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(motorGroup)
I1 = simulator.NewPredicateSatisfactionTime("I1", MotorGroupState.IsOutOfService, False)
I2 = simulator.NewPredicateSatisfactionTime("I2", MotorGroupState.IsMotor1Working, False)
I3 = simulator.NewPredicateSatisfactionTime("I3", MotorGroupState.IsMotor2Working, False)

missionTime = 8760
numberOfTrials = 100000
simulator.MonteCarloSimulation(missionTime, numberOfTrials)

print("I1\t{0:.3e}".format(I1.GetMean()/missionTime))
print("I2\t{0:.3e}".format(I2.GetMean()/missionTime))
print("I3\t{0:.3e}".format(I3.GetMean()/missionTime))

